@@ -1,6 +1,5 @@
+/* patchlevel.h */
 /*
- * patchlevel.h
- *
  * Copyright (C) 2004-2006 by the Massachusetts Institute of Technology.
  * All rights reserved.
  *
@@ -51,8 +50,8 @@
  * organization.
  */
 #define KRB5_MAJOR_RELEASE 1
-#define KRB5_MINOR_RELEASE 9
-#define KRB5_PATCHLEVEL 5
+#define KRB5_MINOR_RELEASE 10
+#define KRB5_PATCHLEVEL 0
 /* #undef KRB5_RELTAIL */
-#define KRB5_RELDATE "20130425"
-#define KRB5_RELTAG "krb5-1.9.5-final"
+#define KRB5_RELDATE "20120127"
+#define KRB5_RELTAG "tags/krb5-1-10-final"
